/*
Module : SQLiteWrappers.h
Purpose: Defines the interface for a set of C++ wrapper classes which encapsulate SQLite v3 (http://www.sqlite.org).
History: PJN / 13-06-2012 1. Initial public release
         PJN / 28-06-2012 1. CValue methods have been made const where appropriate
                          2. CStatement methods have been made const where appropriate
                          3. Implemented a CStatement::ExecRowset method which provides STL iterator accessor to the 
                          returned rowset. The class returned from ExecRowset is SQLite3::CRowset which provides an 
                          input STL iterator to traverse though all the rows returned by SQLite3::CStatement::Step. 
                          Each row as returned from the CRowset iterator is represented by a SQLite3::CRow. Finally 
                          SQLite3::CRow provides a random access STL iterator to all the columns returning a 
                          SQLite3::CValue
                          4. Two additional methods called SQLite3::CDB::ExecRowset and SQLite3::CDB::ExecRowset16 are
                          provided which encapsulate SQLite3::CStatement::ExecRowset which do not require any parameter
                          binding. The output parameter from these two methods is a SQLite3::CRowset2 which contains
                          an SQLite3::CStatement as a member variable.
                          5. Addition of a "SQLITE3WRAPPERS_LOCAL_INCLUSION" pre-processor define which if set will 
                          cause SQLite3 to be pulled in via a #include "SQLite3.h" instead of via #include <SQLite3.h>.
                          This allows SQLiteWrappers to be included in projects which statically compile to SQLite3 
                          from the same directory as your application code.
         PJN / 07-07-2012 1. const and const_reverse iterators have been added to CRow class.
                          2. CRow column iterators now support std::distance
                          3. Fixed a number of /analyze compiler warnings
                          4. Classes now support a SQLITE3WRAPPERS_WCHAR_T_SUPPORT preprocessor define which treats
                          all UTF16 strings as wchar_t* instead of void*. This can help avoid client code needing to
                          do casts where it has a wchar_t data type defined. It also can help find errors where you
                          pass UTF8 or narrow strings to a method which requires a UTF16 string. These checks can be
                          made more useful if you are using the /Zc:wchar_t VC compiler setting.
                          5. CStatement::Bind and Bind16 methods which take UTF8 and UTF16 strings now use a default
                          value for the length parameter
         PJN / 21-07-2012 1. SQLITE3WRAPPERS_WCHAR_T_SUPPORT preprocessor value has been removed and been replaced with 
                          the opposite which is SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT. By default the classes will now 
                          support wchar_t string values unless the new SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT is defined.
                          2. Removed unnecessary reinterpret_cast's from the sample app.
         PJN / 23-02-2013 1. Updated copyright details.
                          2. Added support to the CStatement::Bind* methods to support binding by named parameter. 
                          Thanks to "Prepost" for suggesting this nice addition.
         PJN / 25-03-2016 1. Updated copyright details.
                          2. Updated the code to clean compile in VC 2013 & VC 2015
                          3. Addition of a "SQLITE3WRAPPERS_WIN32_INCLUSION" pre-processor define which if set will
                          cause SQLite3 to be pulled in via a #include <winsqlite\winsqlite3.h>. This allows SQLiteWrappers 
                          to be compiled against the Windows 10 DLL version of SQLite. Please note that to use this you must 
                          use the Windows 10.0.10586.0 SDK. Note that even when you do this you will probably get import errors 
                          trying to link your code. This is because the version of SQLite included in Windows 10 is 3.8.8.3 
                          and the winsqlite3.h header file included in the Windows SDK does not have the correct defines in 
                          it to allow you to specify the __stdcall calling convention when linking to the DLL. You can change
                          the calling convention on the translation unit which includes SQLiteWrappers.h in your 
                          application by using the __stdcall (/Gz) compiler setting and then your code should link properly. 
                          A checkin was done in March 2015 to the SQLite codebase which addresses this and you should expect 
                          to see this fix in the next release of the Windows 10 SDK. For the moment because of this issue, 
                          it is still easiest to just add the SQLite amalgamation files to your Visual Studio project to pull 
                          in support for SQLite3.
                          4. Updated the code to compile against SQLite 3.11.1. Added support for sqlite3_close_v2, 
                          sqlite3_malloc64, sqlite3_realloc64, sqlite3_bind_blob64, sqlite3_bind_text64, 
                          sqlite3_bind_zeroblob64, sqlite3_value_subtype, sqlite3_value_dup, sqlite3_value_free,
                          sqlite3_stmt_scanstatus, sqlite3_stmt_scanstatus_reset, sqlite3_db_cacheflush, 
                          sqlite3_snapshot_get & sqlite3_snapshot_open
         PJN / 18-03-2017 1. Updated copyright details.
                          2. Updated the code to compile against SQLite 3.17.0. Added support for sqlite3_expanded_sql,
                          sqlite3_trace_v2, sqlite3_set_authorizer & sqlite3_system_errno.
                          3. Updated the code to clean compile in VC 2017
                          4. Removed support for the deprecated APIs: sqlite3_trace & sqlite3_profile
                          5. Replaced NULL with nullptr throughout the code. This does mean that VC 2010 is now the
                          minimum compiler required.
         PJN / 29-04-2017 1. Updated the code to compile cleanly using /permissive-.
         PJN / 20-08-2017 1. Updated the code to compile against SQLite 3.20.0. Added support for 
                          sqlite3_set_last_insert_rowid, sqlite3_prepare_v3, sqlite3_prepare16_v3, sqlite3_bind_pointer &
                          sqlite3_value_pointer.
                          2. Updated the demo app to consistently use prefix increment for iterators in for loops.
                          3. Updated the demo app to compile cleanly when SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT is defined.
                          4. Updated the code to compile cleanly with the winsqlite3.h header file in versions of the
                          Windows SDK v10.0.10586, v10.0.14393, v10.0.15063 & v10.0.16232.
         PJN / 28-12-2017 1. Updated the code to compile against SQLite 3.21.0. No additional changes had to be made to 
                          code to support this new version.
                          2. Verified the code compiled cleanly against winsqlite3.h header file in versions of the
                          Windows SDK v10.0.16257, v10.0.16278, v10.0.16288, v10.0.16299, v10.0.17025 & v10.0.17061.
         PJN / 18-03-2018 1. Updated copyright details
                          2. Updated the code to compile against SQLite 3.22.0. Added support for
                          sqlite3_value_nochange.
                          3. Verified the code compiled cleanly against winsqlite3.h header file in versions of the
                          Windows SDK v10.0.17069, v10.0.17083, v10.0.17095 & v10.0.17111.
                          4. Thoroughly reviewed all classes to ensure they had move constructors and move operator= 
                          as necessary
                          5. All asserts in the code are now only active in debug builds
         PJN / 04-06-2018 1. Updated the code to compile against SQLite 3.23.1. Added support for sqlite3_serialize, 
                          sqlite3_deserialize, sqlite3changeset_apply, sqlite3changeset_apply_v2, 
                          sqlite3changeset_apply_strm & sqlite3changeset_apply_v2_strm
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code
                          will now only compile on VC 2017 or later.
                          3. Verified the code compiled cleanly against winsqlite3.h header file in versions of the
                          Windows SDK v10.0.17125 & v10.0.17134.
         PJN / 30-06-2018 1. Verified the code works against SQLite 3.24.0.
         PJN / 29-10-2018 1. Fixed some further C++ core guidelines compiler warnings.
                          2. Updated the code to compile against SQLite 3.25.2. Added support for 
                          sqlite3_create_window_function.
         PJN / 20-01-2019 1. Updated copyright details.
                          2. Updated the code to compile against SQLite 3.26.0. Added support for 
                          sqlite3_normalized_sql.
         PJN / 09-02-2019 1. Verified the code works against SQLite v3.27.1
         PJN / 07-04-2019 1. Fixed some compiler warnings when compiled with Visual Studio 2019
                          2. Verified the code works against SQLite v3.27.2
         PJN / 17-04-2019 1. Updated the code to compile against SQLite 3.28.0. Added support for 
                          sqlite3_stmt_isexplain & sqlite3_value_frombind.
         PJN / 11-07-2019 1. Verified the code works against SQLite v3.29.0
         PJN / 11-10-2019 1. Updated the code to compile against SQLite 3.30.1. Added support for
                          sqlite3_drop_modules.
         PJN / 25-01-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 16-05-2020 1. Fixed compiler warnings when compiled using C++ 17
                          2. Verified the code works against SQLite v3.31.1
                          3. Fixed more Clang-Tidy static code analysis warnings in the code.
                          4. Fixed some problems when compiling against the latest version of winsqlite3.h
         PJN / 24-05-2020 1. Verified the code works against SQLite v3.32.0.
         PJN / 14-06-2020 1. Verified the code works against SQLite v3.32.2.
                          2. Made more methods nodiscard and constexpr.
                          3. Fixed some compile problems when using VS 2019 Preview.
                          4. Fixed a bug in CColumnIterator::operator> and CRowsetIterator::operator>.
         PJN / 05-12-2020 1. Updated the code to compile against SQLite 3.34.0. Added support for
                          sqlite3_txn_state.
         PJN / 30-01-2021 1. Updated copyright details.
                          2. Verified the code works against SQLite v3.34.1
         PJN / 13-03-2021 1. Verified the code works against SQLite v3.35.0
         PJN / 11-05-2021 1. Verified the code works against SQLite v3.35.3
         PJN / 11-07-2021 1. Updated the code to compile against SQLite v3.36.0. Added support for
                          sqlite3_preupdate_hook, sqlite3_preupdate_old, sqlite3_preupdate_count,
                          sqlite3_preupdate_depth, sqlite3_preupdate_new & sqlite3_preupdate_blobwrite
         PJN / 21-01-2022 1. Updated copyright details.
                          2. Updated the code to compile against SQLite v3.37.2. Added support for
                          sqlite3_int64 sqlite3_changes64, sqlite3_total_changes64 &
                          sqlite3_autovacuum_pages.
                          3. Fixed more Clang-Tidy static code analysis warnings in the code.
                          4. Fixed compiler warnings when compiled using Visual Studio 2022.
         PJN / 12-02-2022 1. Updated the code to use C++ uniform initialization for all variable declarations
         PJN / 18-03-2022 1. Updated the code to compile against SQLite v3.38.1. Added support for
                          sqlite3_error_offset, sqlite3_vtab_in_first & sqlite3_vtab_in_next.
         PJN / 16-07-2022 1. Updated the code to compile against SQLite v3.39.1. Added support for
                          sqlite3_db_name.
         PJN / 16-07-2022 1. Updated the code to compile against SQLite v3.40.0. Added support for
                          sqlite3_filename typedef & sqlite3_value_encoding.
         PJN / 24-02-2023 1. Updated copyright details.
                          2. Updated the code to compile against SQLite v3.41.0. Added support for
                          sqlite3_is_interrupted & sqlite3_stmt_scanstatus_v2.
         PJN / 27-08-2023 1. Updated the code to compile against SQLite v3.43.0. Added support for
                          sqlite3_stmt_explain.
         PJN / 02-11-2023 1. Updated the code to compile against SQLite v3.44.0. Added support for
                          sqlite3_get_clientdata, sqlite3_set_clientdata & sqlite3changeset_upgrade.
         PJN / 12-01-2024 1. Updated copyright details.
                          2. Verified the code works against SQLite v3.44.2
         PJN / 19-01-2024 1. Verified the code works against SQLite v3.45.0
         PJN / 05-02-2024 1. Verified the code works against SQLite v3.45.1
         PJN / 28-03-2024 1. Verified the code works against SQLite v3.45.2

Copyright (c) 2012 - 2024 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __SQLITE3WRAPPERS_H__
#define __SQLITE3WRAPPERS_H__

#ifndef CSQLITEWRAPPERS_EXT_CLASS
#define CSQLITEWRAPPERS_EXT_CLASS
#endif //#ifndef CSQLITEWRAPPERS_EXT_CLASS


//////////////////// Includes /////////////////////////////////////////////////

#if defined(SQLITE3WRAPPERS_WIN32_INCLUSION)
#include <winsqlite\winsqlite3.h>
#pragma comment(lib, "winsqlite3.lib") //Automatically link in the Windows SQLite3 dll

#ifndef SQLITE_CALLBACK
#define SQLITE_CALLBACK
#endif //#ifndef SQLITE_CALLBACK

#elif defined(SQLITE3WRAPPERS_LOCAL_INCLUSION)
#include "sqlite3.h"

#ifndef SQLITE_CALLBACK
#define SQLITE_CALLBACK
#endif //#ifndef SQLITE_CALLBACK

#else
#include <sqlite3.h>

#ifndef SQLITE_CALLBACK
#define SQLITE_CALLBACK
#endif //#ifndef SQLITE_CALLBACK

#endif //#ifdef SQLITE3WRAPPERS_WIN32_INCLUSION

#include <cassert>

#ifndef _ITERATOR_
#pragma message("To avoid this message, please put iterator in your pre compiled header (normally stdafx.h)")
#include <iterator>
#endif //#ifndef _ITERATOR_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_


//////////////////// Classes //////////////////////////////////////////////////

namespace SQLite3
{


//Forward declarations
class CDB;


class CValue //Encapsulates a sqlite3_value*
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CValue() noexcept : m_p{nullptr},
                                   m_bFreeRequired{false}
  {
  }

  CValue(const CValue& value) noexcept : CValue{}
  {
    //Only makes sense to copy an object if it is not marked for auto freeing
  #ifdef _DEBUG
    assert(!value.m_bFreeRequired);
  #endif //#ifdef _DEBUG

    Attach(value.m_p, false);
  }

  CValue(CValue&& value) noexcept : CValue{}
  {
    Attach(value.m_p, value.m_bFreeRequired);
    value.Detach();
  }

  _CONSTEXPR17 explicit CValue(sqlite3_value* pValue, bool bFreeRequired = false) noexcept : m_p{pValue},
                                                                                             m_bFreeRequired{bFreeRequired}
  {
  }

  ~CValue() noexcept
  {
  __if_exists(CValue::Free)
  {
    if (m_bFreeRequired && (m_p  != nullptr))
    {
      Free();
      m_bFreeRequired = false;
    }
  } //__if_exists(CValue::Free)

    if (m_p != nullptr)
      m_p = nullptr;
  }

  CValue& operator=(const CValue& value) noexcept
  {
    if (this != &value)
    {
    __if_exists(CValue::Free)
    {
      if (m_bFreeRequired && (m_p != nullptr))
      {
        Free();
        m_bFreeRequired = false;
      }
    } //__if_exists(CValue::Free)

      //Only makes sense to copy an object if it is not marked for auto freeing
    #ifdef _DEBUG
      assert(!value.m_bFreeRequired);
    #endif //#ifdef _DEBUG

      Attach(value.m_p, false);
    }
    return *this;
  }

  CValue& operator=(CValue&& value) noexcept
  {
  __if_exists(CValue::Free)
  {
    if (m_bFreeRequired && (m_p != nullptr))
    {
      Free();
      m_bFreeRequired = false;
    }
  } //__if_exists(CValue::Free)
    Attach(value.m_p, value.m_bFreeRequired);
    value.Detach();
    return *this;
  }

//Methods
  _NODISCARD const void* Blob() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_blob(m_p);
  }

  _NODISCARD int Bytes() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_bytes(m_p);
  }

  _NODISCARD int Bytes16() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_bytes16(m_p);
  }

  _NODISCARD double Double() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_double(m_p);
  }

  _NODISCARD int Int() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_int(m_p);
  }

  _NODISCARD sqlite3_int64 Int64() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_int64(m_p);
  }

  _NODISCARD const unsigned char* Text() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_text(m_p);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* Text16() const noexcept
#else
  _NODISCARD const void* Text16() const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_value_text16(m_p));
  #else
    return sqlite3_value_text16(m_p);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const void* Text16LE() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_text16le(m_p);
  }

  _NODISCARD const void* Text16BE() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_text16be(m_p);
  }

  _NODISCARD int Type() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_type(m_p);
  }

__if_exists(sqlite3_value_subtype)
{
  _NODISCARD unsigned int Subtype() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_subtype(m_p);
  }
} //__if_exists(sqlite3_value_subtype)

  _NODISCARD int NumericType() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_numeric_type(m_p);
  }

__if_exists(sqlite3_value_dup)
{
  _NODISCARD CValue Duplicate() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return CValue(sqlite3_value_dup(m_p), true);
  }
} //__if_exists(sqlite3_value_dup)

__if_exists(sqlite3_value_free)
{
  void Free() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    sqlite3_value_free(m_p);
    m_p = nullptr;
  }
} //__if_exists(sqlite3_value_free)

__if_exists(sqlite3_value_pointer)
{
  _NODISCARD void* Pointer(const char* szType) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_pointer(m_p, szType);
  }
} //__if_exists(sqlite3_value_pointer)

__if_exists(sqlite3_value_nochange)
{
  _NODISCARD int NoChange() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_nochange(m_p);
  }
} //__if_exists(sqlite3_value_nochange)

__if_exists(sqlite3_value_frombind)
{
  _NODISCARD int FromBind() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_value_frombind(m_p);
  }
} //__if_exists(sqlite3_value_frombind)

__if_exists(sqlite3_value_encoding)
{
  _NODISCARD int Encoding() const noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_value_encoding(m_p);
  }
} //__if_exists(sqlite3_value_encoding)

__if_exists(sqlite3_vtab_in_first)
{
  int VTabInFirst(CValue& value) const noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    sqlite3_value* pOut{nullptr};
    const int nReturn{sqlite3_vtab_in_first(m_p, &pOut)};
    value.Attach(pOut, false);
    return nReturn;
  }
} //__if_exists(sqlite3_vtab_in_first)

__if_exists(sqlite3_vtab_in_next)
{
  int VTabInNext(CValue& value) const noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    sqlite3_value* pOut{nullptr};
    const int nReturn{sqlite3_vtab_in_next(m_p, &pOut)};
    value.Attach(pOut, false);
    return nReturn;
  }
} //__if_exists(sqlite3_vtab_in_next)

  operator sqlite3_value*() const noexcept
  {
    return m_p;
  }

  void Attach(sqlite3_value* pValue, bool bFreeRequired) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    if (m_bFreeRequired)
      assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    m_p = pValue;
    m_bFreeRequired = bFreeRequired;
  }

  sqlite3_value* Detach() noexcept
  {
    sqlite3_value* pValue{m_p};
    m_p = nullptr;
    return pValue;
  }

//Member variables
  sqlite3_value* m_p;
  bool m_bFreeRequired;
};


//forward declaration
class CRowset;


class CStatement //Encapsulates a sqlite3_stmt*
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CStatement() noexcept : m_p{nullptr},
                                       m_bAutoFinalize{true}
  {
  }

  _CONSTEXPR17 explicit CStatement(sqlite3_stmt* pStmt, bool bAutoFinalize = true) noexcept : m_p{pStmt},
                                                                                              m_bAutoFinalize{bAutoFinalize}
  {
  }

  CStatement(const CStatement&) = delete;

  CStatement(CStatement&& statement) noexcept : CStatement()
  {
    Attach(statement, statement.m_bAutoFinalize);
    statement.Detach();
  }

  ~CStatement() noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoFinalize)
        Finalize();
      else
        m_p = nullptr;
    }
  }

  CStatement& operator=(const CStatement&) = delete;

  CStatement& operator=(CStatement&& statement) noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoFinalize)
        Finalize();
      else
        m_p = nullptr;
    }
    Attach(statement, statement.m_bAutoFinalize);
    statement.Detach();
    return *this;
  }

//Methods
  int Prepare(const CDB& db, const char* zSql, int nByte = -1, const char** pzTail = nullptr) noexcept;
  int Preparev2(const CDB& db, const char* zSql, int nByte = -1, const char** pzTail = nullptr) noexcept;
  __if_exists(sqlite3_prepare_v3)
  {
    int Preparev3(const CDB& db, const char* zSql, unsigned int prepFlags, int nByte = -1, const char** pzTail = nullptr) noexcept;
  } //__if_exists(sqlite3_prepare_v3)

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int Prepare16(const CDB& db, const wchar_t* zSql, int nByte = -1, const wchar_t** pzTail = nullptr) noexcept;
  int Prepare16v2(const CDB& db, const wchar_t* zSql, int nByte = -1, const wchar_t** pzTail = nullptr) noexcept;
  __if_exists(sqlite3_prepare16_v3)
  {
    int Prepare16v3(const CDB& db, const wchar_t* zSql, unsigned int prepFlags, int nByte = -1, const wchar_t** pzTail = nullptr) noexcept;
  } //__if_exists(sqlite3_prepare16_v3)
#else
  int Prepare16(const CDB& db, const void* zSql, int nByte = -1, const void** pzTail = nullptr) noexcept;
  int Prepare16v2(const CDB& db, const void* zSql, int nByte = -1, const void** pzTail = nullptr) noexcept;
  __if_exists(sqlite3_prepare16_v3)
  {
    int Prepare16v3(const CDB& db, const void* zSql, unsigned int prepFlags, int nByte = -1, const void** pzTail = nullptr) noexcept;
  } //__if_exists(sqlite3_prepare16_v3)
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT

#pragma warning(suppress: 26477)
  int Bind(int nIndex, const void* pData, int n, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_blob(m_p, nIndex, pData, n, pFree);
  }

__if_exists(sqlite3_bind_blob64)
{
#pragma warning(suppress: 26477)
  int Bind64(int nIndex, const void* pData, sqlite3_uint64 n, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_blob64(m_p, nIndex, pData, n, pFree);
  }
} //__if_exists(sqlite3_bind_blob64)

#pragma warning(suppress: 26477)
  int Bind(const char* zName, const void* pData, int n, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, pData, n, pFree);
  }

__if_exists(sqlite3_bind_blob64)
{
#pragma warning(suppress: 26477)
  int Bind64(const char* zName, const void* pData, sqlite3_uint64 n, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind64(nIndex, pData, n, pFree);
  }
} //__if_exists(sqlite3_bind_blob64)

  int Bind(int nIndex, double dblValue) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_double(m_p, nIndex, dblValue);
  }

  int Bind(const char* zName, double dblValue) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, dblValue);
  }

  int Bind(int nIndex, int nValue) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_int(m_p, nIndex, nValue);
  }

  int Bind(const char* zName, int nValue) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, nValue);
  }

  int Bind(int nIndex, sqlite3_int64 nValue) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_int64(m_p, nIndex, nValue);
  }

  int Bind(const char* zName, sqlite3_int64 nValue) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, nValue);
  }

  int BindNULL(int nIndex) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_null(m_p, nIndex);
  }

  int BindNULL(const char* zName) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return BindNULL(nIndex);
  }

#pragma warning(suppress: 26477)
  int Bind(int nIndex, const char* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_text(m_p, nIndex, pData, n, pFree);
  }

__if_exists(sqlite3_bind_text64)
{
  int Bind64(int nIndex, const char* pData, sqlite3_uint64 n, void (SQLITE_CALLBACK *pFree)(void*), unsigned char encoding) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_text64(m_p, nIndex, pData, n, pFree, encoding);
  }
} //__if_exists(sqlite3_bind_text64)

#pragma warning(suppress: 26477)
  int Bind(const char* zName, const char* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, pData, n, pFree);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26477)
  int Bind16(int nIndex, const wchar_t* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
#else
#pragma warning(suppress: 26477)
  int Bind16(int nIndex, const void* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_text16(m_p, nIndex, pData, n, pFree);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26477)
  int Bind16(const char* zName, const wchar_t* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
#else
#pragma warning(suppress: 26477)
  int Bind16(const char* zName, const void* pData, int n = -1, void (SQLITE_CALLBACK *pFree)(void*) = SQLITE_STATIC) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind16(nIndex, pData, n, pFree);
  }

  int Bind(int nIndex, const CValue& value) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_value(m_p, nIndex, value);
  }

  int Bind(const char* zName, const CValue& value) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return Bind(nIndex, value);
  }

  int BindZeroBlob(int nIndex, int n) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_zeroblob(m_p, nIndex, n);
  }

__if_exists(sqlite3_bind_zeroblob64)
{
  int BindZeroBlob64(int nIndex, sqlite3_uint64 n) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_zeroblob64(m_p, nIndex, n);
  }
} //__if_exists(sqlite3_bind_zeroblob64)

  int BindZeroBlob(const char* zName, int n) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return BindZeroBlob(nIndex, n);
  }

__if_exists(sqlite3_bind_zeroblob64)
{
  int BindZeroBlob64(const char* zName, sqlite3_uint64 n) noexcept
  {
    //Convert the name to an index
    const int nIndex{ParameterIndex(zName)};
    if (nIndex == 0)
      return SQLITE_RANGE;

    return BindZeroBlob64(nIndex, n);
  }
} //__if_exists(sqlite3_bind_zeroblob64)

__if_exists(sqlite3_bind_pointer)
{
  int BindPointer(int nIndex, void* pData, const char* szType, void (SQLITE_CALLBACK *pDestructor)(void*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_pointer(m_p, nIndex, pData, szType, pDestructor);
  }
} //__if_exists(sqlite3_bind_pointer)

  int ExecRowset(CRowset& rowset) noexcept;

  int Step() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_step(m_p);
  }

  _NODISCARD int ColumnCount() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_count(m_p);
  }

  _NODISCARD int ParameterCount() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_parameter_count(m_p);
  }

  int ParameterIndex(const char* zName) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_parameter_index(m_p, zName);
  }

  _NODISCARD const char* ParameterName(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_bind_parameter_name(m_p, n);
  }

  int Reset() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_reset(m_p);
  }

  int Finalize() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    const int nReturn{sqlite3_finalize(m_p)};
    m_p = nullptr;
    return nReturn;
  }

  _NODISCARD const char* ColumnDatabaseName(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_database_name(m_p, n);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnDatabaseName16(int n) const noexcept
#else
  _NODISCARD const void* ColumnDatabaseName16(int n) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_database_name16(m_p, n));
  #else
    return sqlite3_column_database_name16(m_p, n);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const char* ColumnTableName(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_table_name(m_p, n);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnTableName16(int n) const noexcept
#else
  _NODISCARD const void* ColumnTableName16(int n) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_table_name16(m_p, n));
  #else
    return sqlite3_column_table_name16(m_p, n);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const char* ColumnOriginName(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_origin_name(m_p, n);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnOriginName16(int n) const noexcept
#else
  _NODISCARD const void* ColumnOriginName16(int n) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_origin_name16(m_p, n));
  #else
    return sqlite3_column_origin_name16(m_p, n);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const char* ColumnDecltype(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_decltype(m_p, n);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnDecltype16(int n) const noexcept
#else
  _NODISCARD const void* ColumnDecltype16(int n) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_decltype16(m_p, n));
  #else
    return sqlite3_column_decltype16(m_p, n);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const char* ColumnName(int n) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_name(m_p, n);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnName16(int n) const noexcept
#else
  _NODISCARD const void* ColumnName16(int n) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_name16(m_p, n));
  #else
    return sqlite3_column_name16(m_p, n);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD const void* ColumnBlob(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_blob(m_p, iCol);
  }

  _NODISCARD int ColumnBytes(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_bytes(m_p, iCol);
  }

  _NODISCARD int ColumnBytes16(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_bytes16(m_p, iCol);
  }

  _NODISCARD double ColumnDouble(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_double(m_p, iCol);
  }

  _NODISCARD int ColumnInt(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_int(m_p, iCol);
  }

  _NODISCARD sqlite3_int64 ColumnInt64(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_int64(m_p, iCol);
  }

  _NODISCARD const unsigned char* ColumnText(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_text(m_p, iCol);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ColumnText16(int iCol) const noexcept
#else
  _NODISCARD const void* ColumnText16(int iCol) const noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_column_text16(m_p, iCol));
  #else
    return sqlite3_column_text16(m_p, iCol);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  _NODISCARD int ColumnType(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_column_type(m_p, iCol);
  }

  _NODISCARD CValue ColumnValue(int iCol) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return CValue{sqlite3_column_value(m_p, iCol), false};
  }

  _NODISCARD int DataCount() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_data_count(m_p);
  }

  int ClearBindings() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_clear_bindings(m_p);
  }

  CDB DB() noexcept;

  _NODISCARD const char* SQL() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_sql(m_p);
  }

__if_exists(sqlite3_expanded_sql)
{
  _NODISCARD char* ExpandedSQL() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_expanded_sql(m_p);
  }
} //__if_exists(sqlite3_expanded_sql)

  _NODISCARD int Busy() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_stmt_busy(m_p);
  }

  _NODISCARD int ReadOnly() const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_stmt_readonly(m_p);
  }

  _NODISCARD int Status(int op, int resetFlg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_stmt_status(m_p, op, resetFlg);
  }

__if_exists(sqlite3_stmt_scanstatus)
{
  int ScanStatus(int idx, int iScanStatusOp, void *pOut) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_stmt_scanstatus(m_p, idx, iScanStatusOp, pOut);
  }
} //__if_exists(sqlite3_stmt_scanstatus)

__if_exists(sqlite3_stmt_scanstatus_v2)
{
  int ScanStatusv2(int idx, int iScanStatusOp, int flags, void* pOut) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_stmt_scanstatus_v2(m_p, idx, iScanStatusOp, flags, pOut);
  }
} //__if_exists(sqlite3_stmt_scanstatus_v2)

__if_exists(sqlite3_stmt_scanstatus_reset)
{
  void ScanStatusReset() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    sqlite3_stmt_scanstatus_reset(m_p);
  }
} //__if_exists(sqlite3_stmt_scanstatus_reset)

__if_exists(sqlite3_normalized_sql)
{
  _NODISCARD const char* NormalizedSQL() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_normalized_sql(m_p);
  }
} //__if_exists(sqlite3_normalized_sql)

__if_exists(sqlite3_stmt_isexplain)
{
  _NODISCARD int IsExplain() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_stmt_isexplain(m_p);
  }
} //__if_exists(sqlite3_stmt_isexplain)

__if_exists(sqlite3_stmt_explain)
{
  int Explain(int eMode) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_stmt_explain(m_p, eMode);
  }
} //__if_exists(sqlite3_stmt_explain)

  operator sqlite3_stmt*() const noexcept
  {
    return m_p;
  }

  void Attach(sqlite3_stmt* pStmt, bool bAutoFinalize = true) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    m_p = pStmt;
    m_bAutoFinalize = bAutoFinalize;
  }

  sqlite3_stmt* Detach() noexcept
  {
    sqlite3_stmt* pStmt{m_p};
    m_p = nullptr;
    m_bAutoFinalize = false;
    return pStmt;
  }

//Member variables
  sqlite3_stmt* m_p;
  bool m_bAutoFinalize;
};


//An STL iterator for CRow
template<class _Ty, class _non_const_Ty, class TRow>
class CColumnIterator
{
public:
//Typedefs
  using iterator_category = std::random_access_iterator_tag;
  using value_type = _Ty;
  using difference_type = ptrdiff_t;
  using pointer = _Ty*;
  using reference = _Ty&;
  using const_self_type = const CColumnIterator;
  using self_type = CColumnIterator;
  using size_type = size_t;

//Constructors / Destructors
  _CONSTEXPR17 CColumnIterator() noexcept : m_pRow{nullptr},
                                            m_nPosition{-1}
  {
  }

  _CONSTEXPR17 CColumnIterator(const_self_type& other) = default;
  _CONSTEXPR17 CColumnIterator(self_type&& other) = default;

  _CONSTEXPR17 CColumnIterator(TRow* pRow, difference_type nPosition) noexcept : m_pRow{pRow},
                                                                                 m_nPosition{nPosition}
  {
  }

  ~CColumnIterator() = default;

//Methods
  _NODISCARD _CONSTEXPR17 reference operator*() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_nPosition != -1);
#endif //#ifdef _DEBUG

#pragma warning(suppress: 26446)
    return operator[](m_nPosition);
  }

  _CONSTEXPR17 self_type& operator=(const_self_type& value) = default;
  _CONSTEXPR17 self_type& operator=(self_type&& value) = default;

  _NODISCARD _CONSTEXPR17 pointer operator->()
  {
    return &(operator*());
  }

  _NODISCARD reference operator[](size_type nPosition) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_pRow != nullptr);
  #endif //#ifdef _DEBUG
    __analysis_assume(m_pRow != nullptr);

#pragma warning(suppress: 26446)
    return m_pRow->operator[](nPosition);
  }

  _CONSTEXPR17 self_type& operator++() noexcept
  {
    ++m_nPosition;
    return *this;
  }

  _CONSTEXPR17 self_type operator++(int)
  {
    self_type tmp{*this};
    ++(*this);
    return tmp;
  }

  _CONSTEXPR17 self_type& operator--() noexcept
  {
    --m_nPosition;
    return *this;
  }

  _CONSTEXPR17 self_type operator--(int)
  {
    self_type tmp{*this};
    --(*this);
    return tmp;
  }

  _NODISCARD _CONSTEXPR17 self_type operator+(const difference_type n)
  {
    self_type tmp{*this};
    tmp.m_nPosition += n;
    return tmp;
  }

  _CONSTEXPR17 self_type& operator+=(const difference_type n) noexcept
  {
    m_nPosition += n;
    return *this;
  }

  _NODISCARD _CONSTEXPR17 self_type operator-(const difference_type n)
  {
    self_type tmp{*this};
    tmp.m_nPosition -= n;
    return tmp;
  }

  _CONSTEXPR17 self_type& operator-=(difference_type n) noexcept
  {
    m_nPosition -= n;
    return *this;
  }

  _NODISCARD _CONSTEXPR17 bool operator==(const_self_type& other) const noexcept
  {
    return m_nPosition == other.m_nPosition;
  }

  _NODISCARD _CONSTEXPR17 bool operator!=(const_self_type& other) const noexcept
  {
    return !operator==(other);
  }

  _NODISCARD _CONSTEXPR17 bool operator<(const_self_type& other) const noexcept
  {
    return m_nPosition < other.m_nPosition;
  }

  _NODISCARD _CONSTEXPR17 bool operator<=(const_self_type& other) const noexcept
  {
    return operator<(other) || operator==(other);
  }

  _NODISCARD _CONSTEXPR17 bool operator>(const_self_type& other) const noexcept
  {
    return !operator<(other) && !operator==(other);
  }

  _NODISCARD _CONSTEXPR17 bool operator>=(const_self_type& other) const noexcept
  {
    return operator>(other) || operator==(other);
  }

  _NODISCARD _CONSTEXPR17 difference_type operator-(const_self_type& other) noexcept
  {
    return m_nPosition - other.m_nPosition;
  }

  _NODISCARD _CONSTEXPR17 friend difference_type operator-(const_self_type& left, const_self_type& right) noexcept
  {
    return left.m_nPosition - right.m_nPosition;
  }

protected:
//Member variables
  TRow* m_pRow;
  difference_type m_nPosition;
};


class CRow //Represents a row as contained in CRowset
{
public:
//Typedefs
  using iterator = CColumnIterator<CValue, CValue, CRow>;
  using const_iterator = CColumnIterator<const CValue, CValue, const CRow>;
  using reverse_iterator = std::reverse_iterator<CColumnIterator<CValue, CValue, CRow> >;
  using const_reverse_iterator = std::reverse_iterator<CColumnIterator<const CValue, CValue, const CRow> >;
  using value_type = CValue;
  using reference = CValue&;
  using const_reference = const CValue&;
  using size_type = size_t;
  using difference_type = iterator::difference_type;

//Constructors / Destructors
  CRow(CStatement* pStatement)  noexcept : m_pStatement{pStatement},
                                           m_nCachedSize{-1}
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_pStatement != nullptr);
  #endif //#ifdef _DEBUG
    __analysis_assume(m_pStatement != nullptr);

    //Cache the column count
    m_nCachedSize = m_pStatement->ColumnCount();
  }

//Methods
#pragma warning(suppress: 26487)
  _CONSTEXPR17 iterator begin() noexcept
  {
    return {this, 0};
  }

  _CONSTEXPR17 const_iterator begin() const noexcept
  {
    return {this, 0};
  }

  _CONSTEXPR17 const_iterator cbegin() const noexcept
  {
    return begin();
  }

#pragma warning(suppress: 26487)
  _CONSTEXPR17 iterator end() noexcept
  {
#pragma warning(suppress: 26472)
    return {this, static_cast<iterator::difference_type>(size())};
  }

  _CONSTEXPR17 const_iterator end() const noexcept
  {
#pragma warning(suppress: 26472)
    return {this, static_cast<iterator::difference_type>(size())};
  }

  _CONSTEXPR17 const_iterator cend() const noexcept
  {
    return end();
  }

#pragma warning(suppress: 26440)
  _CONSTEXPR17 reverse_iterator rbegin()
  {
#pragma warning(suppress: 26487)
    return reverse_iterator{end()};
  }

#pragma warning(suppress: 26440)
  _CONSTEXPR17 const_reverse_iterator rbegin() const
  {
#pragma warning(suppress: 26487)
    return const_reverse_iterator{end()};
  }

  _CONSTEXPR17 const_reverse_iterator crbegin() const
  {
    return rbegin();
  }

#pragma warning(suppress: 26440)
  _CONSTEXPR17 reverse_iterator rend()
  {
#pragma warning(suppress: 26487)
    return reverse_iterator{begin()};
  }

#pragma warning(suppress: 26440)
  _CONSTEXPR17 const_reverse_iterator rend() const
  {
#pragma warning(suppress: 26487)
    return const_reverse_iterator{begin()};
  }

  _CONSTEXPR17 const_reverse_iterator crend() const
  {
    return rend();
  }

  reference operator[](size_type nPosition) const noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_pStatement != nullptr);
    assert(nPosition < size());
  #endif //#ifdef _DEBUG
    __analysis_assume(m_pStatement != nullptr);

    //Update the cached value
#pragma warning(suppress: 26472)
    m_Value = m_pStatement->ColumnValue(static_cast<int>(nPosition));

    //Return the cached value
    return m_Value;
  }

  _CONSTEXPR17 size_type size() const noexcept
  {
    return m_nCachedSize;
  }

  _CONSTEXPR17 bool empty() const noexcept
  {
    return size() == 0;
  }

//Member variables
  CStatement* m_pStatement;
  mutable value_type m_Value;
  int m_nCachedSize;
#pragma warning(suppress: 26495)
};


//An STL iterator for CRowset
class CRowsetIterator
{
public:
//Typedefs
  using iterator_category = std::input_iterator_tag;
  using value_type = CRow;
  using difference_type = ptrdiff_t;
  using pointer = CRow*;
  using reference = CRow&;
  using const_self_type = const CRowsetIterator;
  using self_type = CRowsetIterator;
  using size_type = size_t;
  using const_pointer = const CRow*;
  using const_reference = const CRow&;

//Constructors / Destructors
  CRowsetIterator() = delete;
  CRowsetIterator(CRowset* pRowset, sqlite3_int64 nPosition) noexcept;
  CRowsetIterator(const_self_type& iterator) = default;
  CRowsetIterator(self_type&& iterator) = default;
#pragma warning(suppress: 26487)
  ~CRowsetIterator() = default;

//Methods
#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 reference operator*() noexcept
  {
    return m_Row;
  }

  self_type& operator=(const_self_type& value) = default;
  self_type& operator=(self_type && value) = default;

  _NODISCARD _CONSTEXPR17 pointer operator->() noexcept
  {
    return &(operator*());
  }

  self_type& operator++() noexcept;

  self_type operator++(int) noexcept
  {
#pragma warning(suppress: 26487)
    const self_type tmp{*this};
    ++(*this);
#pragma warning(suppress: 26486)
    return tmp;
  }

#pragma warning(suppress: 26487)
  _NODISCARD self_type operator+(const difference_type n) noexcept
  {
    self_type tmp{*this};
    for (difference_type i=0; i<n; i++)
      ++(tmp);
    return tmp;
  }

#pragma warning(suppress: 26487)
  self_type& operator+=(const difference_type n) noexcept
  {
    for (difference_type i=0; i<n; i++)
      operator++();
    return *this;
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator==(const self_type& other) const noexcept
  {
    return m_nPosition == other.m_nPosition;
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator!=(const self_type& other) const noexcept
  {
    return !operator==(other);
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator<(const self_type& other) const noexcept
  {
    if ((m_nPosition == -1) && (other.m_nPosition == -1))
      return false;
    else if (m_nPosition == -1)
      return false;
    else if (other.m_nPosition == -1)
      return true;
    else
      return m_nPosition < other.m_nPosition;
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator<=(const self_type& other) const noexcept
  {
    return operator<(other) || operator==(other);
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator>(const self_type& other) const noexcept
  {
    return !operator<(other) && !operator==(other);
  }

#pragma warning(suppress: 26487)
  _NODISCARD _CONSTEXPR17 bool operator>=(const self_type& other) const noexcept
  {
    return operator>(other) || operator==(other);
  }

protected:
//Member variables
  CRowset* m_pRowset;
  value_type m_Row;
  sqlite3_int64 m_nPosition;
};


class CRowset //Encapsulates a rowset as returned from CStatement::ExecRowset
{
public:
//Typedefs
  using iterator = CRowsetIterator;

//Constructors / Destructors
  _CONSTEXPR17 CRowset() noexcept : m_nFirstStepReturnValue{SQLITE_ERROR},
                                    m_pStatement{nullptr}
  {
  }

//Methods
  _NODISCARD iterator begin() noexcept
  {
    if (m_nFirstStepReturnValue == SQLITE_ROW)
      return {this, 0};
    else
      return end();
  }

  _NODISCARD iterator end() noexcept
  {
    return {this, -1};
  }

protected:
//Member variables
  int m_nFirstStepReturnValue;
  CStatement* m_pStatement;

  friend class CRowsetIterator;
  friend class CStatement;
};


class CValue2 : public CValue //version of CValue with an embedded CStatement. Used by the CDB::ExecScalar[16] methods
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CValue2()  noexcept : CValue{}
  {
  }

  CValue2(const CValue& value) noexcept : CValue{}
  {
    //Only makes sense to copy an object if it is not marked for auto freeing
  #ifdef _DEBUG
    assert(!value.m_bFreeRequired);
  #endif //#ifdef _DEBUG

    //Let the base class do its thing
    __super::operator=(value);
  }

//Methods
#pragma warning(suppress: 26434)
  CValue2& operator=(const CValue& value) noexcept
  {
    if (this != &value)
      __super::operator=(value);
    return *this;
  }

//Member variables
  CStatement m_Statement;
};


class CRowset2 : public CRowset //version of CRowset with an embedded CStatement. Used by the CDB::ExecRowset[16] methods
{
public:
//Member variables
  CStatement m_Statement;
};


class CDB //Encapsulates a sqlite3*
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CDB() noexcept : m_p{nullptr},
                                m_bAutoClose{false}
  {
  }

  explicit CDB(sqlite3* pDB, bool bAutoClose = true)  noexcept : m_p{nullptr},
                                                                 m_bAutoClose{bAutoClose}
  {
    Attach(pDB, bAutoClose);
  }

  CDB(CDB&& db) noexcept : CDB{}
  {
    Attach(db.m_p, db.m_bAutoClose);
    db.Detach();
  }

  CDB(const CDB&) = delete;

  ~CDB() noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoClose)
        Close();
      else
        m_p = nullptr;
    }
  }

  CDB& operator=(CDB&& db) noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoClose)
        Close();
      else
        m_p = nullptr;
    }
    Attach(db.m_p, db.m_bAutoClose);
    db.Detach();
    return *this;
  }

  CDB& operator=(const CDB&) = delete;

//Methods
  int Open(const char* filename) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_open(filename, &m_p);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int Open16(const wchar_t* filename) noexcept
#else
  int Open16(const void* filename) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_open16(filename, &m_p);
  }

  int Open(const char* filename, int flags, const char* pszVFS) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_open_v2(filename, &m_p, flags, pszVFS);
  }

  int Close() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    const int nReturn{sqlite3_close(m_p)};
    m_p = nullptr;
    return nReturn;
  }

__if_exists(sqlite3_close_v2)
{
  int Closev2() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    const int nReturn{sqlite3_close_v2(m_p)};
    m_p = nullptr;
    return nReturn;
  }
} //__if_exists(sqlite3_close_v2)

  _NODISCARD int Changes() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_changes(m_p);
  }

__if_exists(sqlite3_changes64)
{
  _NODISCARD sqlite3_int64 Changes64() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_changes64(m_p);
  }
} //__if_exists(sqlite3_changes64)

  _NODISCARD int TotalChanges() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_total_changes(m_p);
  }

__if_exists(sqlite3_total_changes64)
{
  _NODISCARD sqlite3_int64 TotalChanges64() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_total_changes64(m_p);
  }
} //__if_exists(sqlite3_total_changes64)

  int BusyHandler(int(SQLITE_CALLBACK *pBusyHandler)(void*, int), void* pData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_busy_handler(m_p, pBusyHandler, pData);
  }

  int BusyTimeout(int ms) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_busy_timeout(m_p, ms);
  }

  int CollationNeeded(void* pData, void (SQLITE_CALLBACK *pCallback)(void*, sqlite3*, int eTextRep, const char*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_collation_needed(m_p, pData, pCallback);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int CollationNeeded16(void* pData, void (SQLITE_CALLBACK *pCallback)(void*, sqlite3*, int eTextRep, const wchar_t*)) noexcept
#else
  int CollationNeeded16(void* pData, void (SQLITE_CALLBACK *pCallback)(void*, sqlite3*, int eTextRep, const void*)) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26490)
    return sqlite3_collation_needed16(m_p, pData, reinterpret_cast<void (SQLITE_CALLBACK *)(void*, sqlite3*, int, const void*)>(pCallback));
  #else
    return sqlite3_collation_needed16(m_p, pData, pCallback);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  int CreateCollation(const char* zName, int eTextRep, void* pArg, int (SQLITE_CALLBACK *pCompare)(void*, int, const void*, int, const void*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_collation(m_p, zName, eTextRep, pArg, pCompare);
  }

  int CreateCollation(const char* zName, int eTextRep, void* pArg, int (SQLITE_CALLBACK *pCompare)(void*, int, const void*, int, const void*), void (SQLITE_CALLBACK *pDestroy)(void*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_collation_v2(m_p, zName, eTextRep, pArg, pCompare, pDestroy);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int CreateCollation16(const wchar_t* zName, int eTextRep, void* pArg, int(SQLITE_CALLBACK *pCompare)(void*, int, const void*, int, const void*)) noexcept
#else
  int CreateCollation16(const void* zName, int eTextRep, void* pArg, int(SQLITE_CALLBACK *pCompare)(void*, int, const void*, int, const void*)) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_collation16(m_p, zName, eTextRep, pArg, pCompare);
  }

  int CreateFunction(const char* zFunctionName, int nArg, int eTextRep, void* pApp, void (SQLITE_CALLBACK *pFunc)(sqlite3_context*, int, sqlite3_value**),
                     void (SQLITE_CALLBACK *pStep)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *pFinal)(sqlite3_context*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_function(m_p, zFunctionName, nArg, eTextRep, pApp, pFunc, pStep, pFinal);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int CreateFunction16(const wchar_t* zFunctionName, int nArg, int eTextRep, void* pApp, void (SQLITE_CALLBACK *pFunc)(sqlite3_context*, int, sqlite3_value**),
                       void (SQLITE_CALLBACK *pStep)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *pFinal)(sqlite3_context*)) noexcept
#else
  int CreateFunction16(const void* zFunctionName, int nArg, int eTextRep, void* pApp, void (SQLITE_CALLBACK *pFunc)(sqlite3_context*, int, sqlite3_value**),
                       void (SQLITE_CALLBACK *pStep)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *pFinal)(sqlite3_context*)) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_function16(m_p, zFunctionName, nArg, eTextRep, pApp, pFunc, pStep, pFinal);
  }

  int CreateFunction(const char* zFunctionName, int nArg, int eTextRep, void* pApp, void (SQLITE_CALLBACK *pFunc)(sqlite3_context*, int, sqlite3_value**),
                     void (SQLITE_CALLBACK *pStep)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *pFinal)(sqlite3_context*), void (SQLITE_CALLBACK *pDestroy)(void*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_function_v2(m_p, zFunctionName, nArg, eTextRep, pApp, pFunc, pStep, pFinal, pDestroy);
  }

  int ExtendedResultCodes(int onoff) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_extended_result_codes(m_p, onoff);
  }

  int FileControl(const char* zDbName, int op, void* pData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_file_control(m_p, zDbName, op, pData);
  }

  int GetAutocommit() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_get_autocommit(m_p);
  }

  bool NextStatement(CStatement& statement) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    sqlite3_stmt* pStatement{sqlite3_next_stmt(m_p, statement)};

    statement.Detach();
    statement.Attach(pStatement, false);

    return (pStatement != nullptr);
  }

  void Interrupt() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_interrupt(m_p);
  }

__if_exists(sqlite3_is_interrupted)
{
  int IsInterrupted() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_is_interrupted(m_p);
  }
} //__if_exists(sqlite3_is_interrupted)

  _NODISCARD sqlite3_int64 LastInsertRowID() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_last_insert_rowid(m_p);
  }

__if_exists(sqlite3_set_last_insert_rowid)
{
  void SetLastInsertRowID(sqlite3_int64 nValue) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    sqlite3_set_last_insert_rowid(m_p, nValue);
  }
} //__if_exists(sqlite3_set_last_insert_rowid)

  int Limit(int id, int newVal) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_limit(m_p, id, newVal);
  }

  int EnableLoadExtension(int onoff) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_enable_load_extension(m_p, onoff);
  }

  int LoadExtension(const char* zFile, const char* zProc, char** pzErrMsg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_load_extension(m_p, zFile, zProc, pzErrMsg);
  }

  void* CommitHook(int(SQLITE_CALLBACK *pCallback)(void*), void* pData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_commit_hook(m_p, pCallback, pData);
  }

  void* RollbackHook(void (SQLITE_CALLBACK *pCallback)(void*), void* pData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_rollback_hook(m_p, pCallback, pData);
  }

  _NODISCARD int ErrorCode() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_errcode(m_p);
  }

  _NODISCARD int ExtendedErrorCode() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_extended_errcode(m_p);
  }

  _NODISCARD const char* ErrorMessage() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_errmsg(m_p);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  _NODISCARD const wchar_t* ErrorMessage16() noexcept
#else
  _NODISCARD const void* ErrorMessage16() noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
    return static_cast<const wchar_t*>(sqlite3_errmsg16(m_p));
  #else
    return sqlite3_errmsg16(m_p);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }

  int Exec(const char* sql, int(SQLITE_CALLBACK *pCallback)(void*, int, char**, char**) = nullptr, void* pData = nullptr, char** errmsg = nullptr) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_exec(m_p, sql, pCallback, pData, errmsg);
  }

  int ExecScalar(const char* sql, CValue2& value, int nByte = -1, const char** pzTail = nullptr) noexcept
  {
    //Prepare the statement
    if (value.m_Statement.operator sqlite3_stmt*() != nullptr)
      value.m_Statement.Finalize();
    int nRet{value.m_Statement.Preparev2(*this, sql, nByte, pzTail)};
    if (nRet != SQLITE_OK)
      return nRet;

    //Check we have got back at least one column
    if (value.m_Statement.ColumnCount() < 1)
      return SQLITE_ERROR;

    //Move to the first row of the recordset
    nRet = value.m_Statement.Step();
    if (nRet == SQLITE_ROW)
      value = value.m_Statement.ColumnValue(0);
    else
      nRet = SQLITE_ERROR;

    return nRet;
  }

__if_exists(CStatement::Preparev3)
{
  int ExecScalarv3(const char* sql, CValue2& value, unsigned int prepFlags, int nByte = -1, const char** pzTail = nullptr) noexcept
  {
    //Prepare the statement
    if (value.m_Statement.operator sqlite3_stmt*() != nullptr)
      value.m_Statement.Finalize();
    int nRet{value.m_Statement.Preparev3(*this, sql, prepFlags, nByte, pzTail)};
    if (nRet != SQLITE_OK)
      return nRet;

    //Check we have got back at least one column
    if (value.m_Statement.ColumnCount() < 1)
      return SQLITE_ERROR;

    //Move to the first row of the recordset
    nRet = value.m_Statement.Step();
    if (nRet == SQLITE_ROW)
      value = value.m_Statement.ColumnValue(0);
    else
      nRet = SQLITE_ERROR;

    return nRet;
  }
} //__if_exists(CStatement::Preparev3)

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int Exec16Scalar(const wchar_t* sql, CValue2& value) noexcept
#else
  int Exec16Scalar(const void* sql, CValue2& value) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Prepare the statement
    if (value.m_Statement.operator sqlite3_stmt*() != nullptr)
      value.m_Statement.Finalize();
    int nRet{value.m_Statement.Prepare16v2(*this, sql)};
    if (nRet != SQLITE_OK)
      return nRet;

    //Check we have got back at least one column
    if (value.m_Statement.ColumnCount() < 1)
      return SQLITE_ERROR;

    //Move to the first row of the recordset
    nRet = value.m_Statement.Step();
    if (nRet == SQLITE_ROW)
      value = value.m_Statement.ColumnValue(0);
    else
      nRet = SQLITE_ERROR;

    return nRet;
  }

  int ExecRowset(const char* sql, CRowset2& rowset) noexcept
  {
    //Prepare the statement
    if (rowset.m_Statement.operator sqlite3_stmt*() != nullptr)
      rowset.m_Statement.Finalize();
    const int nRet{rowset.m_Statement.Preparev2(*this, sql)};
    if (nRet != SQLITE_OK)
      return nRet;

    //Move to the first row of the recordset
    return rowset.m_Statement.ExecRowset(rowset);
  }

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  int ExecRowset16(const wchar_t* sql, CRowset2& rowset) noexcept
#else
  int ExecRowset16(const void* sql, CRowset2& rowset) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Prepare the statement
    if (rowset.m_Statement.operator sqlite3_stmt*() != nullptr)
      rowset.m_Statement.Finalize();
    const int nRet{rowset.m_Statement.Prepare16v2(*this, sql)};
    if (nRet != SQLITE_OK)
      return nRet;

    //Move to the first row of the recordset
    return rowset.m_Statement.ExecRowset(rowset);
  }

  int OverloadFunction(const char* zFuncName, int nArg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_overload_function(m_p, zFuncName, nArg);
  }

__if_exists(sqlite3_trace_v2)
{
  int Trace(unsigned uMask, int(SQLITE_CALLBACK *xCallback)(unsigned, void*, void*, void*), void* pCtx) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_trace_v2(m_p, uMask, xCallback, pCtx);
  }
} //__if_exists(sqlite3_trace_v2)

  void ProgressHandler(int n, int(SQLITE_CALLBACK *pCallback)(void*), void* pData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    sqlite3_progress_handler(m_p, n, pCallback, pData);
  }

  int CreateModule(const char* zName, const sqlite3_module* p, void* pClientData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_module(m_p, zName, p, pClientData);
  }

  int CreateModule(const char* zName, const sqlite3_module* p, void* pClientData, void (SQLITE_CALLBACK *pDestroy)(void*)) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_module_v2(m_p, zName, p, pClientData, pDestroy);
  }

  sqlite3_filename DBFilename(const char* zDbName) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_filename(m_p, zDbName);
  }

  _NODISCARD sqlite3_mutex* DBMutex() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_mutex(m_p);
  }

  int DBReleaseMemory() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_release_memory(m_p);
  }

  int DBStatus(int op, int* pCur, int* pHiwtr, int resetFlg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_status(m_p, op, pCur, pHiwtr, resetFlg);
  }

  int DeclareVirtualTable(const char* zSQL) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_declare_vtab(m_p, zSQL);
  }

  int DBReadOnly(const char* zDbName) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_readonly(m_p, zDbName);
  }

  int TableColumnMetadata(const char* zDbName, const char* zTableName, const char* zColumnName, const char** pzDataType, 
                          const char** pzCollSeq, int* pNotNull, int* pPrimaryKey, int* pAutoinc) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_table_column_metadata(m_p, zDbName, zTableName, zColumnName, pzDataType, pzCollSeq, pNotNull, pPrimaryKey, pAutoinc);
  }

  int UnlockNotify(void (SQLITE_CALLBACK *pNotify)(void** apArg, int nArg), void* pNotifyArg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_unlock_notify(m_p, pNotify, pNotifyArg);
  }

  void* UpdateHook(void (SQLITE_CALLBACK *pHook)(void*, int , char const* ,char const*, sqlite3_int64), void* pHookArg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_update_hook(m_p, pHook, pHookArg);
  }

  int VirtualTableOnConflict() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_vtab_on_conflict(m_p);
  }

  int WALAutoCheckpoint(int N) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_wal_autocheckpoint(m_p, N);
  }

  int WALCheckpoint(const char* zDb) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_wal_checkpoint(m_p, zDb);
  }

  int WALCheckpoint(const char* zDb, int eMode, int* pnLog = nullptr, int* pnCkpt = nullptr) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_wal_checkpoint_v2(m_p, zDb, eMode, pnLog, pnCkpt);
  }

  void* WALHook(int(SQLITE_CALLBACK *pHook)(void*, sqlite3*, const char* , int), void* pHookArg) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_wal_hook(m_p, pHook, pHookArg);
  }

__if_exists(sqlite3_db_cacheflush)
{
  int CacheFlush() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_db_cacheflush(m_p);
  }
} //__if_exists(sqlite3_db_cacheflush)

__if_exists(sqlite3_snapshot_get)
{
  int SnapshotGet(const char* zSchema, sqlite3_snapshot** ppSnapshot) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_snapshot_get(m_p, zSchema, ppSnapshot);
  }
} //__if_exists(sqlite3_snapshot_get)

__if_exists(sqlite3_snapshot_open)
{
  int SnapshotOpen(const char* zSchema, sqlite3_snapshot* pSnapshot) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_snapshot_open(m_p, zSchema, pSnapshot);
  }
} //__if_exists(sqlite3_snapshot_open)

__if_exists(sqlite3_set_authorizer)
{
  int SetAuthorizer(int(SQLITE_CALLBACK *xAuth)(void*, int, const char*, const char*, const char*, const char*), void* pUserData) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_set_authorizer(m_p, xAuth, pUserData);
  }
} //__if_exists(sqlite3_set_authorizer)

//Note unfortunately we cannot wrap sqlite3_db_config because as of v3.37.2 there is no version
//of sqlite3_db_config which takes a va_list parameter

//Note unfortunately we cannot wrap sqlite3_vtab_config because as of v3.37.2 there is no version
//of sqlite3_vtab_config which takes a va_list parameter

__if_exists(sqlite3_system_errno)
{
  int SystemErrno() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_system_errno(m_p);
  }
} //__if_exists(sqlite3_system_errno)

__if_exists(sqlite3_serialize)
{
  unsigned char* Serialize(const char* zSchema, sqlite3_int64* piSize, unsigned int mFlags) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_serialize(m_p, zSchema, piSize, mFlags);
  }
} //__if_exists(sqlite3_serialize)

__if_exists(sqlite3_deserialize)
{
  int Deserialize(const char* zSchema, unsigned char* pData, sqlite3_int64 szDb, sqlite3_int64 szBuf, unsigned mFlags) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_deserialize(m_p, zSchema, pData, szDb, szBuf, mFlags);
  }
} //__if_exists(sqlite3_deserialize)

__if_exists(sqlite3changeset_apply)
{
  int ChangesetApply(int nChangeset, void* pChangeset, int (*xFilter)(void *pCtx, const char* zTab), int (*xConflict)(void* pCtx, int eConflict, sqlite3_changeset_iter* p), void* pCtx) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3changeset_apply(m_p, nChangeset, pChangeset, xFilter, xConflict, pCtx);
  }
} //__if_exists(sqlite3changeset_apply)

__if_exists(sqlite3changeset_apply_v2)
{
  int ChangesetApplyv2(int nChangeset, void* pChangeset, int (*xFilter)(void *pCtx, const char* zTab), int (*xConflict)(void* pCtx, int eConflict, sqlite3_changeset_iter* p), void* pCtx, void** ppRebase, int* pnRebase, int flags) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3changeset_apply_v2(m_p, nChangeset, pChangeset, xFilter, xConflict, pCtx, ppRebase, pnRebase, flags);
  }
} //__if_exists(sqlite3changeset_apply_v2)

__if_exists(sqlite3changeset_apply_strm)
{
  int ChangesetApplyStrm(int (*xInput)(void* pIn, void* pData, int* pnData), void* pIn, int (*xFilter)(void* pCtx, const char* zTab), int (*xConflict)(void* pCtx, int eConflict, sqlite3_changeset_iter* p), void* pCtx) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3changeset_apply_strm(m_p, xInput, pIn, xFilter, xConflict, pCtx);
  }
} //__if_exists(sqlite3changeset_apply_strm)

__if_exists(sqlite3changeset_apply_v2_strm)
{
  int ChangesetApplyStrmv2(int (*xInput)(void* pIn, void* pData, int* pnData), void* pIn, int (*xFilter)(void* pCtx, const char* zTab), int (*xConflict)(void* pCtx, int eConflict, sqlite3_changeset_iter* p), void* pCtx, void** ppRebase, int* pnRebase, int flags) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3changeset_apply_v2_strm(m_p, xInput, pIn, xFilter, xConflict, pCtx, ppRebase, pnRebase, flags);
  }
} //__if_exists(sqlite3changeset_apply_v2_strm)

__if_exists(sqlite3_create_window_function)
{

#ifdef SQLITE3WRAPPERS_WIN32_INCLUSION
  int CreateWindowFunction(const char* zFunctionName, int nArg, int eTextRep, void* pApp, void (SQLITE_CALLBACK *xStep)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *xFinal)(sqlite3_context*), void (SQLITE_CALLBACK *xValue)(sqlite3_context*), void (SQLITE_CALLBACK *xInverse)(sqlite3_context*, int, sqlite3_value**), void (SQLITE_CALLBACK *xDestroy)(void*)) noexcept
#else
  int CreateWindowFunction(const char* zFunctionName, int nArg, int eTextRep, void* pApp, void (*xStep)(sqlite3_context*, int, sqlite3_value**), void (*xFinal)(sqlite3_context*), void (*xValue)(sqlite3_context*), void (*xInverse)(sqlite3_context*, int, sqlite3_value**), void (*xDestroy)(void*)) noexcept
#endif //ifdef SQLITE3WRAPPERS_WIN32_INCLUSION
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_create_window_function(m_p, zFunctionName, nArg, eTextRep, pApp, xStep, xFinal, xValue, xInverse, xDestroy);
  }
} //__if_exists(sqlite3_create_window_function)

__if_exists(sqlite3_drop_modules)
{
  int DropModules(const char** azKeep) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_drop_modules(m_p, azKeep);
  }
} //__if_exists(sqlite3_drop_modules)

__if_exists(sqlite3_txn_state)
{
  int TxnState(const char* zSchema) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_txn_state(m_p, zSchema);
  }
} //__if_exists(sqlite3_txn_state)

__if_exists(sqlite3_preupdate_hook)
{
  void* PreupdateHook(void (*xPreUpdate)(void* pCtx, sqlite3* db, int op, char const* zDb, char const* zName, sqlite3_int64 iKey1, sqlite3_int64 iKey2), void* pParam) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_hook(m_p, xPreUpdate, pParam);
  }
} //__if_exists(sqlite3_preupdate_hook)

__if_exists(sqlite3_preupdate_old)
{
  int PreupdateOld(int nValue, sqlite3_value** ppValue) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_old(m_p, nValue, ppValue);
  }
} //__if_exists(sqlite3_preupdate_old)

__if_exists(sqlite3_preupdate_count)
{
  int PreupdateCount() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_count(m_p);
  }
} //__if_exists(sqlite3_preupdate_count)

__if_exists(sqlite3_preupdate_depth)
{
  int PreupdateDepth() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_depth(m_p);
  }
} //__if_exists(sqlite3_preupdate_depth)

__if_exists(sqlite3_preupdate_new)
{
  int PreupdateNew(int nValue, sqlite3_value** ppValue) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_new(m_p, nValue, ppValue);
  }
} //__if_exists(sqlite3_preupdate_new)

__if_exists(sqlite3_preupdate_blobwrite)
{
  int PreupdateBlobWrite() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_preupdate_blobwrite(m_p);
  }
} //__if_exists(sqlite3_preupdate_blobwrite)

__if_exists(sqlite3_autovacuum_pages)
{
  int AutoVacuumPages(unsigned int (*C)(void*, const char*, unsigned int, unsigned int, unsigned int), void* P, void(*X)(void*)) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_autovacuum_pages(m_p, C, P, X);
  }
} //__if_exists(sqlite3_autovacuum_pages)

__if_exists(sqlite3_error_offset)
{
  _NODISCARD int ErrorOffset() noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_error_offset(m_p);
  }
} //__if_exists(sqlite3_error_offset)

__if_exists(sqlite3_db_name)
{
  _NODISCARD const char* DBName(int N) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_db_name(m_p, N);
  }
} //__if_exists(sqlite3_db_name)

__if_exists(sqlite3_get_clientdata)
{
  _NODISCARD void* GetClientData(const char* N) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_get_clientdata(m_p, N);
  }
} //__if_exists(sqlite3_get_clientdata)

__if_exists(sqlite3_set_clientdata)
{
  int SetClientData(const char* N, void* P, void(*X)(void*)) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3_set_clientdata(m_p, N, P, X);
  }
} //__if_exists(sqlite3_set_clientdata)

__if_exists(sqlite3changeset_upgrade)
{
  int ChangesetUpgrade(const char* zDb, int nIn, const void* pIn, int* pnOut, void** ppOut) noexcept
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_p != nullptr);
#endif //#ifdef _DEBUG

    return sqlite3changeset_upgrade(m_p, zDb, nIn, pIn, pnOut, ppOut);
  }
} //__if_exists(sqlite3changeset_upgrade)

  operator sqlite3*() const noexcept
  {
    return m_p;
  }

  void Attach(sqlite3* pDB, bool bAutoClose = true) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG
    m_p = pDB;
    m_bAutoClose = bAutoClose;
  }

  sqlite3* Detach() noexcept
  {
    sqlite3* pDB{m_p};
    m_p = nullptr;
    m_bAutoClose = false;
    return pDB;
  }

//Member variables
  sqlite3* m_p;
  bool m_bAutoClose;
};


class CBlob //Encapsulates a sqlite3_blob*
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CBlob() noexcept : m_p{nullptr},
                                  m_bAutoClose{false}
  {
  }

  _CONSTEXPR17 explicit CBlob(sqlite3_blob* pBlob, bool bAutoClose = true) noexcept : m_p{pBlob},
                                                                                      m_bAutoClose{bAutoClose}
  {
  }

  CBlob(const CBlob&) = delete;

  CBlob(CBlob&& blob) noexcept : CBlob{}
  {
    Attach(blob.m_p, blob.m_bAutoClose);
    blob.Detach();
  }

  ~CBlob() noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoClose)
        Close();
      else
        m_p = nullptr;
    }
  }

  CBlob& operator=(const CBlob&) = delete;

  CBlob& operator=(CBlob&& blob) noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoClose)
        Close();
      else
        m_p = nullptr;
    }
    Attach(blob.m_p, blob.m_bAutoClose);
    return *this;
  }

//Methods
  int Open(const CDB& db, const char* zDb, const char* zTable, const char* zColumn, sqlite3_int64 iRow, int flags) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_open(db, zDb, zTable, zColumn, iRow, flags, &m_p);
  }

  int Close() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_close(m_p);
  }

  _NODISCARD int Bytes() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_bytes(m_p);
  }

  int Read(void* Z, int N, int iOffset) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_read(m_p, Z, N, iOffset);
  }

  int Reopen(sqlite3_int64 iRow) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_reopen(m_p, iRow);
  }

  int Write(const void* z, int n, int iOffset) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_blob_write(m_p, z, n, iOffset);
  }

  operator sqlite3_blob*() const noexcept
  {
    return m_p;
  }

  void Attach(sqlite3_blob* pBlob, bool bAutoClose = true) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    m_p = pBlob;
    m_bAutoClose = bAutoClose;
  }

  sqlite3_blob* Detach() noexcept
  {
    sqlite3_blob* pBlob{m_p};
    m_p = nullptr;
    m_bAutoClose = false;
    return pBlob;
  }

//Member variables
  sqlite3_blob* m_p;
  bool m_bAutoClose;
};


class CPtr //Encapsulates a pointer which needs to be deallocated using sqlite3_free as well as SQLite printf functionality
{
public:
//Constructors / Destructors
  _CONSTEXPR17 CPtr() noexcept : m_p{nullptr},
                                 m_bAutoFree{false}
  {
  }

  CPtr(const CPtr&) = delete;

  CPtr(CPtr&& ptr) noexcept : CPtr{}
  {
    Attach(ptr.m_p, ptr.m_bAutoFree);
    ptr.Detach();
  }

  ~CPtr() noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoFree)
        Free();
      else
        m_p = nullptr;
    }
  }

  CPtr& operator=(const CPtr&) = delete;

  CPtr& operator=(CPtr&& ptr) noexcept
  {
    if (m_p != nullptr)
    {
      if (m_bAutoFree)
        Free();
      else
        m_p = nullptr;
    }
    Attach(ptr.m_p, ptr.m_bAutoFree);
    ptr.Detach();
    return *this;
  }

//Methods
  bool Allocate(int nBytes) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
    assert(nBytes != 0); //Does not make sense to allocate zero bytes
  #endif //#ifdef _DEBUG

    m_p = sqlite3_malloc(nBytes);
    return (m_p != nullptr);
  }

__if_exists(sqlite3_malloc64)
{
  bool Allocate64(sqlite3_uint64 nBytes) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
    assert(nBytes != 0); //Does not make sense to allocate zero bytes
  #endif //#ifdef _DEBUG

    m_p = sqlite3_malloc64(nBytes);
    return (m_p != nullptr);
  }
} //__if_exists(sqlite3_malloc64)

  bool Reallocate(int nBytes) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
    assert(nBytes != 0); //Does not make sense to reallocate zero bytes
  #endif //#ifdef _DEBUG

    m_p = sqlite3_realloc(m_p, nBytes);
    return (m_p != nullptr);
  }

__if_exists(sqlite3_realloc64)
{
  bool Reallocate64(sqlite3_uint64 nBytes) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
    assert(nBytes != 0); //Does not make sense to reallocate zero bytes
  #endif //#ifdef _DEBUG

    m_p = sqlite3_realloc64(m_p, nBytes);
    return (m_p != nullptr);
  }
} //__if_exists(sqlite3_realloc64)

  void Free() noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p != nullptr);
    assert(m_bAutoFree);
  #endif //#ifdef _DEBUG

    sqlite3_free(m_p);
    m_p = nullptr;
    m_bAutoFree = false;
  }

  bool Format(const char* zFormat, ...) noexcept
  {
    if (m_p != nullptr)
      Free();

#pragma warning(suppress: 26826)
    va_list argp;
#pragma warning(suppress: 26492 26481 26826)
    va_start(argp, zFormat);
#pragma warning(suppress: 26486)
    m_p = sqlite3_vmprintf(zFormat, argp);
#pragma warning(suppress: 26477 26826)
    va_end(argp);
    return (m_p != nullptr);
  }

  operator void*() const noexcept
  {
    return m_p;
  }

  operator const char*() const noexcept
  {
    return static_cast<const char*>(m_p);
  }

  void Attach(void* pPtr, bool bAutoFree = true) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
  #endif //#ifdef _DEBUG

    m_p = pPtr;
    m_bAutoFree = bAutoFree;
  }

  void* Detach() noexcept
  {
    void* pPtr{m_p};
    m_p = nullptr;
    return pPtr;
  }

//Member variables
  void* m_p;
  bool m_bAutoFree;
};


inline CRowsetIterator::CRowsetIterator(CRowset* pRowset, sqlite3_int64 nPosition) noexcept : m_pRowset{pRowset},
                                                                                              m_Row{pRowset->m_pStatement},
                                                                                              m_nPosition{nPosition}
{
}

#pragma warning(suppress: 26487)
inline CRowsetIterator::self_type& CRowsetIterator::operator++() noexcept
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_pRowset != nullptr);
  assert(m_pRowset->m_pStatement != nullptr);
#endif //#ifdef _DEBUG
  __analysis_assume(m_pRowset != nullptr);
  __analysis_assume(m_pRowset->m_pStatement != nullptr);
  const int nRet{m_pRowset->m_pStatement->Step()};
#ifdef _DEBUG
  assert((nRet == SQLITE_DONE) || (nRet == SQLITE_ROW));
#endif //#ifdef _DEBUG

  ++m_nPosition;
  if (nRet == SQLITE_DONE)
    m_nPosition = -1;
  return *this;
}

inline CDB CStatement::DB() noexcept
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_p != nullptr);
#endif //#ifdef _DEBUG

  return CDB(sqlite3_db_handle(m_p), false);
}

inline int CStatement::Prepare(const CDB& db, const char* zSql, int nByte, const char** pzTail) noexcept
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_p == nullptr);
  assert(db.operator sqlite3*() != nullptr);
#endif //#ifdef _DEBUG

  return sqlite3_prepare(db, zSql, nByte, &m_p, pzTail);
}

inline int CStatement::Preparev2(const CDB& db, const char* zSql, int nByte, const char** pzTail) noexcept
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_p == nullptr);
  assert(db.operator sqlite3*() != nullptr);
#endif //#ifdef _DEBUG

  return sqlite3_prepare_v2(db, zSql, nByte, &m_p, pzTail);
}

__if_exists(sqlite3_prepare_v3)
{
  inline int CStatement::Preparev3(const CDB& db, const char* zSql, unsigned int prepFlags, int nByte, const char** pzTail) noexcept
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
    assert(db.operator sqlite3*() != nullptr);
  #endif //#ifdef _DEBUG

    return sqlite3_prepare_v3(db, zSql, nByte, prepFlags, &m_p, pzTail);
  }
} //__if_exists(sqlite3_prepare_v3)

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
inline int CStatement::Prepare16(const CDB& db, const wchar_t* zSql, int nByte, const wchar_t** pzTail) noexcept
#else
inline int CStatement::Prepare16(const CDB& db, const void* zSql, int nByte, const void** pzTail) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_p == nullptr);
  assert(db.operator sqlite3*() != nullptr);
#endif //#ifdef _DEBUG

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26490)
  return sqlite3_prepare16(db, zSql, nByte, &m_p, reinterpret_cast<const void**>(pzTail));
#else
  return sqlite3_prepare16(db, zSql, nByte, &m_p, pzTail);
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
}

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
inline int CStatement::Prepare16v2(const CDB& db, const wchar_t* zSql, int nByte, const wchar_t** pzTail) noexcept
#else
inline int CStatement::Prepare16v2(const CDB& db, const void* zSql, int nByte, const void** pzTail) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
{
  //Validate our parameters
#ifdef _DEBUG
  assert(m_p == nullptr);
  assert(db.operator sqlite3*() != nullptr);
#endif //#ifdef _DEBUG

#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26490)
  return sqlite3_prepare16_v2(db, zSql, nByte, &m_p, reinterpret_cast<const void**>(pzTail));
#else
  return sqlite3_prepare16_v2(db, zSql, nByte, &m_p, pzTail);
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
}

__if_exists(sqlite3_prepare16_v3)
{
#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  inline int CStatement::Prepare16v3(const CDB& db, const wchar_t* zSql, unsigned int prepFlags, int nByte, const wchar_t** pzTail) noexcept
#else
  inline int CStatement::Prepare16v3(const CDB& db, const void* zSql, unsigned int prepFlags, int nByte, const void** pzTail) noexcept
#endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_p == nullptr);
    assert(db.operator sqlite3*() != nullptr);
  #endif //#ifdef _DEBUG

  #ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
#pragma warning(suppress: 26490)
    return sqlite3_prepare16_v3(db, zSql, nByte, prepFlags, &m_p, reinterpret_cast<const void**>(pzTail));
  #else
    return sqlite3_prepare16_v3(db, zSql, nByte, prepFlags, &m_p, pzTail);
  #endif //#ifndef SQLITE3WRAPPERS_NO_WCHAR_T_SUPPORT
  }
} //__if_exists(sqlite3_prepare16_v3)

inline int CStatement::ExecRowset(CRowset& rowset) noexcept
{
  //Execute the first step
  rowset.m_nFirstStepReturnValue = Step();

  //Update the other values in the rowset instance being returned
  rowset.m_pStatement = this;

  return rowset.m_nFirstStepReturnValue;
}


}; //namespace SQLite3

#endif //#ifndef __SQLITE3WRAPPERS_H__
